#ifndef COSMOLOGY_H
#define COSMOLOGY_H

#include "Common.h"
#include "Math.h"
#include "array.h"

/******************************************************************************
 * Cosmology
 *
 * Contains all information needed to specify a cosmology (for our purposes).
 ******************************************************************************/
struct Cosmology {
    Cosmology();
    Cosmology(real h, real Tcmb, real n, real Omega_m, real Omega_b);
    Cosmology(real h, real Tcmb, real n, real Omega_m, real Omega_b, real sigma8, const array& ki, const array& Ti);
    Cosmology(const char* name);

    void Initialize(real h, real Tcmb, real n, real Omega_m, real Omega_b);
    void Initialize(const char* name);
    void SetTransferFunction(const array& ki, const array& Ti);
    void LoadTransferFunction(const char* tkfile, int kcol = 1, int tcol = 2);
    void NormalizeTransferFunction(real sigma8);
    void CalculateAdditionalParameters();

    /* Input parameters */
    real h;                     // Hubble parameter today
    real Tcmb;                  // CMB temperature today
    real n;                     // scalar spectral index
    real Omega_m;               // matter density parameter today
    real Omega_b;               // baryon density parameter today
    real sigma8;                // power spectrum variance smoothed at 8 Mpc/h
    array ki;                   // wavenumbers for transfer function (h/Mpc)
    array Ti;                   // transfer function

    /* Calculated parameters */
    real H0;                    // Hubble parameter today
    real rho_crit;              // critical density today
    real Omega_gamma;           // photon density parameter
    real Omega_nu;              // neutrino density parameter
    real Omega_r;               // radiation (photon + neutrino) density parameter
    real Omega_Lambda;          // dark energy density parameter
    real a_eq;                  // scale factor at equality
    real z_eq;                  // redshift of equality
    real k_eq;
    real delta_H;               // normalization of linear power spectrum at z = 0

    real H(real a) const {
        return H0 * sqrt(Omega_m/pow3(a) + Omega_Lambda);
    }

    real dHda(real a) const {
        return -3*pow2(H0)*Omega_m/(2*H(a)*pow4(a));
    }

    /* Print cosmology to stdout */
    void Print(const char* prefix = "") const;
};

#endif // COSMOLOGY_H
